;;; -*- lexical-binding: t; -*-

;;; git-find-file.el --- Incremental search for files in a git repo
;;
;; Description: Modelled on Command-T from TextMate (I think).
;;
;; Author: Mark Triggs <mark@dishevelled.net>

;; This file is free software; you can redistribute it and/or modify
;; it under the terms of the GNU General Public License as published by
;; the Free Software Foundation; either version 2, or (at your option)
;; any later version.

;; This file is distributed in the hope that it will be useful,
;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;; GNU General Public License for more details.

;; You should have received a copy of the GNU General Public License
;; along with GNU Emacs; see the file COPYING.  If not, write to
;; the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
;; Boston, MA 02111-1307, USA.

;; Using it:

;;      (define-key global-map (kbd "C-x C-g") 'git-find-file)

;; Then 'C-x C-g' from within your project directory.  You'll see a list of the
;; files in your current Git repository, and anything you type will be used as
;; input to narrow the list.  A pattern like:
;;
;;      => srcmarcmaclj
;;
;; Could match files like:
;;
;;      src/marcgrep/sources/marc_file.clj
;;      src/marcgrep/sources/marcxml_file.clj
;;      src/marcgrep/destinations/marcfile.clj
;;
;; That is, the pattern is applied (very!) fuzzily, and any path where the
;; letters of the pattern appear in order will be returned.
;;
;; There's some basic relevance ranking here: exact substring matches sort above
;; fuzzy matches, fuzzy matches against the path's basename are better than
;; matches against the entire path, and matches where the letters appear close
;; to one another are better than when they're spread out.
;;
;; Hitting RET opens the file under point, and C-g will abort the file selection.
;;
;; Finally, C-s "rotates" the list:
;;
;;      /path/a              /path/b              /path/c
;;      /path/b  --[C-s]-->  /path/c  --[C-s]-->  /path/a
;;      /path/c              /path/a              /path/b
;;

;;; Code:

(require 'cl)

(defvar gff-large-project-threshold 2000
  "Projects with more than this many files will delay updating the file list until `gff-input-delay' seconds have elapsed with no input.")

;;; 200ms seems as low as this will go while still working.  Might need to revisit how I'm grabbing input.
(defvar gff-input-delay 0.2)

(defvar gff-ignored-regexp nil)


(defun gff-start-of-match (pattern input)
  "If the letters of `pattern' appear in order in `input', return
the position in the string of where they start."
  (cl-block nil
    (let ((len (length pattern)))
      (let ((pattern-pos (1- len))
            (input-pos (length input))
            (letter-distance 0)
            (letter-distances ()))
        (while (>= pattern-pos 0)
          (let ((next (position (elt pattern pattern-pos) input
                                :from-end t
                                :end input-pos)))
            (if (not next)
                (cl-return (cons nil nil))
              (cl-incf letter-distance (- input-pos next))
              (cl-decf pattern-pos)
              (setq input-pos next))))
        ;; Refine letter distance by searching forward again
        (let ((pos input-pos)
              (gap-count 0)
              (refined-letter-distance 0))
          (dotimes (i (length pattern))
            (let ((next-pos (position (elt pattern i) input :start pos)))
              (unless (zerop (- next-pos pos))
                (incf gap-count))
              (cl-incf refined-letter-distance (1+ (- next-pos pos)))
              (setq pos (1+ next-pos))))
          (setq letter-distance refined-letter-distance)
          (when (= 1 gap-count)
            ;; Special case: our input sequence is two substrings like somefilebit.java
            ;; Nearly as good as an exact match
            (setq letter-distance (1+ (length pattern)))))
        (cons input-pos letter-distance)))))


(defun gff-score-fuzzily (pattern input)
  (cl-destructuring-bind (match . letter-distance)
      (gff-start-of-match pattern input)
    (when match
      ;; (let* ((basename-start (or (position ?/ input :from-end t)
      ;;                            match))
      ;;        (raw-score (cond ((= (1- match) basename-start) 650)
      ;;                         ((> match basename-start)
      ;;                          (if (string-match "[[:punct:][:space:]]" (string (elt input (1- match))))
      ;;                              600
      ;;                            500))
      ;;                         (t 300))))
      ;;   ;; Penalise results where the letters are some distance apart
      ;;   (- raw-score (* letter-distance 2)))
      (/ 1.0 (/ letter-distance (float (length pattern)))))))


(defun gff-scorers-for (pattern)
  "Build a list of scorer functions that will match `pattern'"
  (list
   `(lambda (s)
      (let* ((basename (file-name-nondirectory s))
            (position-in-basename (search ,pattern basename)))
        (cond ((equal position-in-basename 0) 1100)
              ((equal position-in-basename (- (length basename) ,(length pattern))) 1200)
              (position-in-basename 1000)
              (t nil))))
   `(lambda (s)
      (gff-score-fuzzily ,pattern s))))


(defun find-file-in-parent-dir (target dir)
  (let ((dir (expand-file-name dir)))
    (when (not (string= "/" dir))
      (let ((filename (format "%s%s"
                              (file-name-as-directory dir)
                              target)))
        (if (file-exists-p filename)
            filename
          (find-file-in-parent-dir target
                                   (file-name-directory
                                    (directory-file-name dir))))))))

(defvar gff-old-window-configuration nil)

(defun gff-calculate-starting-directory (n default-dir toplevel-dir)
  (let ((number-of-leading-dirs (length (split-string (file-truename (expand-file-name toplevel-dir)) "/" t))))
    (concat "/"
            (mapconcat 'identity
                       (let ((bits (split-string (file-truename (expand-file-name default-dir)) "/" t)))
                         (subseq bits 0 (min (+ number-of-leading-dirs n) (length bits))))
                       "/")
            "/")))



(defun git-find-file ()
  "Run an interactive search for all files in this repo."
  (interactive)
  (let* ((starting-directory default-directory)
         (default-directory
           (file-truename
            (file-name-directory
             (expand-file-name
              (or (find-file-in-parent-dir ".git" default-directory)
                  (error "No .git directory found!")))))))
    (gff-init (if (numberp current-prefix-arg)
                  (gff-calculate-starting-directory current-prefix-arg starting-directory default-directory)
                default-directory))))


(defun gff-update-filter (fn)
  "Apply `fn' to the current filter and replace it with the result."
  (setq gff-active-filter (funcall fn gff-active-filter)
        gff-rotation 0)
  (message "=> %s" gff-active-filter)
  (gff-enqueue-refresh))


(defvar gff-refresh-pending nil)

(defun gff-enqueue-refresh ()
  "Refresh the buffer once user input seems to have stopped"
  (let ((delay (if (> gff-list-size gff-large-project-threshold)
                   gff-input-delay
                 0)))
    (setq gff-refresh-pending t)
    (when gff-keypress-timer
      (cancel-timer gff-keypress-timer))
    (setq gff-keypress-timer (run-at-time delay nil 'gff-refresh-buffer))))


(defun gff-wait-for-refresh-buffer ()
  (while gff-refresh-pending
    (sit-for 0.01)))


(defun gff-keypress ()
  "Record a keypress"
  (interactive)
  (gff-update-filter `(lambda (s) (concat s ,(string last-command-event)))))


(defun gff-yank ()
  (interactive)
  (gff-update-filter `(lambda (s) (concat s (current-kill 0 t)))))


(defun gff-reset ()
  "Set the current filter string back to empty."
  (interactive)
  (gff-update-filter (lambda (s) "")))


(defun gff-backspace ()
  "Remove the last character from the current filter string."
  (interactive)
  (gff-update-filter (lambda (s)
                       (if (> (length s) 0)
                           (subseq s 0 (1- (length s)))
                         s))))

(defun gff-exit ()
  "Bail out."
  (interactive)
  (kill-buffer nil)
  (when gff-old-window-configuration
    (set-window-configuration (car gff-old-window-configuration))
    (goto-char (or (marker-position (cadr gff-old-window-configuration)) 0))
    (setq gff-old-window-configuration nil)))


(defun gff-rotate-list ()
  "Move the first element of the result list to the end."
  (interactive)
  (cl-incf gff-rotation)
  (gff-refresh-buffer))


(defun gff-get-selection ()
  (let ((result (buffer-substring (line-beginning-position) (line-end-position))))
    (if (equalp result gff-no-match-marker)
        (error "No file selected")
      result)))

(defun gff-select ()
  "Select the file at point"
  (interactive)
  (gff-wait-for-refresh-buffer)
  (let ((dir default-directory)
        (selection (gff-get-selection))
        (handle-fn gff-handle-select-fn)
        (gff-buffer (current-buffer))
        (new-buffer nil))
    (let ((default-directory dir))
      (funcall handle-fn selection)
      (setq new-buffer (current-buffer)))
      (with-current-buffer gff-buffer
        (if gff-suppress-quit
            (setq gff-suppress-quit nil)
          (gff-exit)
          (when (buffer-live-p new-buffer)
            (switch-to-buffer new-buffer))))))


(defun gff-select-dir ()
  "Select the directory of the file at point"
  (interactive)
  (gff-wait-for-refresh-buffer)
  (let ((dir default-directory)
        (selection (gff-get-selection)))
    (gff-exit)
    (let ((default-directory dir))
      (find-file (or (file-name-directory selection) ".")))))


(defun gff-pop-to-buffer (buffer)
  (select-window (split-window-vertically))
  (switch-to-buffer buffer))


(defun gff-init (base-directory &optional buffer-name score-fn list-files-fn handle-select-fn tiebreak-mode init-keymap)
  "Initialise the *git-find-file* buffer to display `files'."
  (setq gff-old-window-configuration (list (current-window-configuration) (point-marker)))
  (let ((buffer-name (or buffer-name "*git-find-file*"))
        (score-fn (or score-fn 'gff-scorers-for)))
    (gff-pop-to-buffer (get-buffer-create buffer-name))

    (if (and current-prefix-arg (listp current-prefix-arg))
      (setq gff-ls-files-command (read-from-minibuffer "List files command?: " "git ls-files"))
      ;; weird setq to make the dynamic scoped version buffer local so it sticks around...
      (setq gff-ls-files-command gff-ls-files-command))

    (buffer-disable-undo nil)

    (set (make-local-variable 'gff-base-directory)
         base-directory)

    (set (make-local-variable 'gff-tiebreak-mode) tiebreak-mode)

    (set (make-local-variable 'gff-list-files-fn) (or list-files-fn
                                                      'gff-git-list-files))

    (set (make-local-variable 'gff-rotation) 0)
    (set (make-local-variable 'gff-former-filters) ())
    (set (make-local-variable 'gff-active-filter) "")
    (set (make-local-variable 'gff-score-function) score-fn)
    (set (make-local-variable 'gff-handle-select-fn)
         (or handle-select-fn 'find-file))

    (set (make-local-variable 'gff-keypress-timer) nil)

    (set (make-local-variable 'gff-suppress-quit) nil)


    (let ((map (make-sparse-keymap)))
      (cl-loop for i from 32 to 127 do (define-key map (string i) 'gff-keypress))
      (cl-loop for i from 160 to 256 do (define-key map (string i) 'gff-keypress))
      (define-key map (kbd "DEL") 'gff-backspace)
      (define-key map (kbd "TAB") 'gff-nested-search)
      (define-key map (kbd "RET") 'gff-select)
      (define-key map (kbd "M-RET") 'gff-select-dir)
      (define-key map [C-return] 'gff-select-dir)
      (define-key map (kbd "C-y") 'gff-yank)
      (define-key map (kbd "C-g") 'gff-exit)
      (define-key map (kbd "C-u") 'gff-reset)
      (define-key map (kbd "C-s") 'isearch-forward)

      (let ((combined (copy-keymap (or init-keymap (make-sparse-keymap)))))
        (set-keymap-parent combined map)
        (use-local-map combined)))

    (gff-refresh-buffer)))

(defun gff-result-sorter (r1 r2)
  ;; Higher score wins
  (cond ((> (car r1) (car r2)) t)
        ((< (car r1) (car r2)) nil)
        (t
         (if (eq gff-tiebreak-mode 'original-order)
             nil
           ;; shortest wins by default
           (< (length (cdr r1)) (length (cdr r2))))
         )))

(defun gff-score (elt scorers)
  (dolist (scorer scorers)
    (let ((score (funcall scorer elt)))
      (when score
        (cl-return score)))))

(defun gff-filter-list (pattern list)
  "Find matches for `pattern' in `list' and return an ordered result set"
  (if (equal pattern "")
      list
    (let* ((scorers (funcall gff-score-function (downcase pattern)))
           (scored-results (cl-loop for elt in list
                                    with score
                                    do (setq score (or (gff-score (downcase elt) scorers)
                                                       0))
                                    if (plusp score)
                                    collect (cons score elt)))
           (sorted-results (stable-sort scored-results 'gff-result-sorter)))
      (mapcar 'cdr sorted-results))))

(defun gff-nested-search ()
  (interactive)
  (push gff-active-filter gff-former-filters)
  (setq gff-active-filter ""))


(defvar-local gff-ls-files-command "git ls-files")

(defun gff-git-list-files ()
  (let* ((fastfilter (executable-find "gff-fastfilter"))
         (filter-cmd (if fastfilter
                         (format "%s %s" fastfilter (mapconcat (lambda (s) (format "'%s'" s)) (cons gff-active-filter gff-former-filters) " "))
                       (mapconcat (lambda (pattern)
                                    (format "egrep -a -i %s"
                                            (shell-quote-argument (if (string= pattern "")
                                                                      "."
                                                                    (mapconcat 'regexp-quote (split-string pattern "" t) ".*")))))
                                  (cons gff-active-filter gff-former-filters)
                                  " | "))))
    (gff-shell-command-to-lines
                   (format "%s %s%s | %s"
                           gff-ls-files-command
                           gff-base-directory
                           (if gff-ignored-regexp
                               (format "| egrep -a -v '%s'" gff-ignored-regexp)
                             "")
                           filter-cmd))))

(defun gff-shell-command-to-lines (command)
  (let ((proc (start-process "sh" nil "sh" "-c" command))
        (partial-lines ())
        (result ())
        (done nil))
    (set-process-filter proc (lambda (proc chunk)
                               (let (eol-index
                                     (offset 0))
                                 (while (setq eol-index (cl-position 10 chunk :start offset))
                                   (push (concat (apply 'concat partial-lines) (subseq chunk offset eol-index))
                                         result)
                                   (setq partial-lines nil)
                                   (setq offset (1+ eol-index)))
                                 (when (< offset (length chunk))
                                   (push (subseq chunk offset) partial-lines)))))
    (while (process-live-p proc)
      (let ((inhibit-quit nil))
       (accept-process-output proc 0.1)))
    (when partial-lines
      (push (apply 'concat partial-lines) result))
    (nreverse result)))


(defvar gff-no-match-marker "(no matches)")

(defun gff-refresh-buffer ()
  "Refresh the *git-find-file* buffer to show the current filtered list of files."
  (when (boundp 'gff-base-directory)
    (let* ((file-list (gff-filter-list gff-active-filter
                                       (funcall gff-list-files-fn)))
           (len (length file-list))
           (inhibit-modification-hooks t))

      (set (make-local-variable 'gff-list-size) len)

      (erase-buffer)
      (if (zerop len)
          (insert gff-no-match-marker)
        (let ((rot 0))
          (dolist (line file-list)
            (if (>= rot gff-rotation)
                (insert line "\n")
              (cl-incf rot)))))
      (goto-char (point-min))))
  (setq gff-refresh-pending nil))


(provide 'git-find-file)
